// Filters.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "filters.h"
#include "ProgressDlg.h"
#include <stdlib.h>
#include <time.h>

#define Period(X) (X - floor(X / M_2PI) * M_2PI)

//     
int freqs[4][6] = 
{{0, 50, 170, 300, 600, 1000},
{5512, 1500, 2000, 3000, 4000, 5000},
{11025,  3000, 5000, 7000, 9000, 11000},
{22050, 3000, 6000, 12000, 14000, 16000}};

//      

int smp_rates[] = {8000, 11025, 22050, 32000, 44100, 48000};

int percentage = 0;
CProgressDlg *pd;
int tf;

void ForwardFFT(int N, double *re, double *im)
{
	memset(im, 0, sizeof(double) * N);
	FFT(N, re, im);
}

void InverseFFT(int N, double *re, double *im)
{
	int i = 0;
	
	for (; i < N; i++)
		im[i] = -im[i];

	FFT(N, re, im);

	for (i = 0; i < N; i++)
	{
		re[i] /= N;
		im[i] /= -N;
	}
}

void FFT(int N, double *re, double *im)
{
	double tr, ti, ur, ui, sr, si;
	int nm1 = N - 1, nd2 = N >> 1, m = (int)ceil(log(N) / log(2));
	int i, j = nd2, k, l, le, le2, jm1, ip;

	for (i = 1; i < nm1; i++)
	{
		if (i < j)
		{
			tr = re[j];
			ti = im[j];
			re[j] = re[i];
			im[j] = im[i];
			re[i] = tr;
			im[i] = ti;
		}
		k = nd2;
		while (k <= j)
		{
			j -= k;
			k >>= 1;
		}
		j += k;
	}

	for (l = 1; l <= m; l++)
	{
		le = 1 << l;
		le2 = le >> 1;
		ur = 1.0;
		ui = 0.0;
		sr = cos(PI / le2);
		si = -sin(PI / le2);
		for (j = 1; j <= le2; j++)
		{
			jm1 = j - 1;
			for (i = jm1; i <= nm1; i += le)
			{
				ip = i + le2;
				tr = re[ip] * ur - im[ip] * ui;
				ti = re[ip] * ui + im[ip] * ur;
				re[ip] = re[i] - tr;
				im[ip] = im[i] - ti;
				re[i] += tr;
				im[i] += ti;
			}
			tr = ur;
			ur = tr * sr - ui * si;
			ui = tr * si + ui * sr;
		}
	}
}

void DFT(int N, double *re, double *im)
{
	double *re_x, *im_x, sr, si;
	double c = 2 * PI / N;
	int i, k;

	re_x = new double[N];
	im_x = new double[N];
	memset(re_x, 0, sizeof(double) * N);
	memset(im_x, 0, sizeof(double) * N);

	for (k = 0; k < N; k++)
		for (i = 0; i < N; i++)
		{
			sr = cos(c * k * i);
			si = -sin(c * k * i);
			re_x[k] += re[i] * sr - im[i] * si;
			im_x[k] += re[i] * si + im[i] * sr;
		}

	memcpy(re, re_x, sizeof(double) * N);
	memcpy(im, im_x, sizeof(double) * N);

	delete im_x;
	delete re_x;
}

void Convolution(int n1, double *x, int n2, double *h, double *y)
{
	int i, j, k, N = n1 + n2 - 1;

	double *t = new double[N];
	memset(t, 0, sizeof(double) * N);

	for (i = 0; i < N; i++)
		for (j = 0; j < n2; j++)
		{
			k = i - j;
			if (k >= 0 && k < n1)
				t[i] += h[j] * x[k];
		}

	memcpy(y, t, sizeof(double) * N);
	delete t;
}

void FFTConvolution(int n1, double *x, int n2, double *h, double *y)
{
	int n = n1 + n2 - 1, i;
	double *re_x, *im_x, *re_h, *im_h, tr, ti;

	n = 1 << (int)ceil(log(n) / log(2));
	re_x = new double[n];
	im_x = new double[n];
	re_h = new double[n];
	im_h = new double[n];

	memcpy(re_x, x, sizeof(double) * n1);
	memset(re_x + n1, 0, sizeof(double) * (n - n1)); 
	memcpy(re_h, h, sizeof(double) * n2);
	memset(re_h + n2, 0, sizeof(double) * (n - n2)); 

	ForwardFFT(n, re_x, im_x);
	ForwardFFT(n, re_h, im_h);

	for (i = 0; i < n; i++)
	{
		tr = re_x[i];
		ti = im_x[i];
		re_x[i] = re_h[i] * tr - im_h[i] * ti;
		im_x[i] = re_h[i] * ti + im_h[i] * tr;
	}

	InverseFFT(n, re_x, im_x);
	memcpy(y, re_x, sizeof(double) * (n1 + n2 - 1));

	delete im_h;
	delete re_h;
	delete im_x;
	delete re_x;
}

double Hamming(int M, int i)
{
	return 0.54 - 0.46 * cos(2 * PI * i / M);
}

double Blackman(int M, int i)
{
	return 0.42 - 0.5 * cos(2 * PI * i / M) +
		0.08 * cos(4 * PI * i / M);
}

void LowPassFIRKernel(int M, double fc, bool IsHamming, double *kernel)
{
	double K = 0.0, c = 2 * PI * fc;
	int m2 = M >> 1, i;

	for (i = 0; i <= M; i++)
	{
		if (!(i - m2))
			kernel[i] = c;
		else
			kernel[i] = sin(c * (i - m2)) / (i - m2);
		if (IsHamming)
			kernel[i] *= Hamming(M, i);
		else
			kernel[i] *= Blackman(M, i);
		K += kernel[i];
	}

	if (K)
	for (i = 0; i <= M; i++)
		kernel[i] /= K;
}

void HighPassFIRKernel(int M, double fc, bool IsHamming, double *kernel)
{
	LowPassFIRKernel(M, fc, IsHamming, kernel);

	for (int i = 0; i <= M; i++)
		kernel[i] = -kernel[i];

	kernel[M >> 1]++;
}

void BandPassFIRKernel(int M, double fc1, double fc2, bool IsHamming, double *kernel)
{
	double *buf = new double[M + 1];

	if (fc1 > fc2)
	{
		double t = fc1;
		fc1 = fc2;
		fc2 = t;
	}

	LowPassFIRKernel(M, fc1, IsHamming, buf);
	LowPassFIRKernel(M, fc2, IsHamming, kernel);

	for (int i = 0; i <= M; i++)
		kernel[i] -= buf[i];

	delete buf;
}

void BandRejectFIRKernel(int M, double fc1, double fc2, bool IsHamming, double *kernel)
{
	double *buf = new double[M + 1];

	if (fc1 > fc2)
	{
		double t = fc1;
		fc1 = fc2;
		fc2 = t;
	}

	LowPassFIRKernel(M, fc1, IsHamming, buf);
	HighPassFIRKernel(M, fc2, IsHamming, kernel);

	for (int i = 0; i <= M; i++)
		kernel[i] += buf[i];

	delete buf;
}

void UserFIRKernel(int M, int *bands, int *freqs, int SmpRate, bool IsHamming, double *kernel)
{
	double *buf = new double[M + 1], n;
	int fc1 = 0, fc2, i = 0;

	memset(kernel, 0, sizeof(double) * (M + 1));

	for (i = 0; i < 10; i++)
	{
		fc2 = freqs[i];
		if (fc1)
			BandPassFIRKernel(M, (double)fc1 / SmpRate, 
				(double)fc2 / SmpRate, IsHamming, buf);
		else
			LowPassFIRKernel(M, (double)fc2 / SmpRate, IsHamming, buf);
		n = pow(10, (double)bands[i] / 20.0);
		for (int j = 0; j <= M; j++)
			kernel[j] += buf[j] * n;
		fc1 = fc2;
	}

	delete buf;
}

void GetPairCoeff(double fc, bool IsLopass, double ripple, int p, int poles, double& a0, double& a1, double& a2, double& b1, double& b2)
{
	double C = M_PI / (poles << 1);
	double C2 = C + (p - 1) * M_PI / poles;
	double rp = -cos(C2);
	double ip = sin(C2);
	double es, vx, kx, t, w, m, d, k, x0, x1, x2, y1, y2;

	if (ripple)
	{
		C = 100.0 / (100.0 - ripple);
		es = sqrt(C * C - 1);
		C = 1.0 / poles;
		C2 = 1.0 / es;
		vx = C * log(C2 + sqrt(C2 * C2 + 1));
		kx = C * log(C2 + sqrt(C2 * C2 - 1));
		kx = cosh(kx);
		rp *= sinh(vx) / kx;
		ip *= cosh(vx) / kx;
	}

	t = 2.0 * tan(0.5);
	w = 2.0 * M_PI * fc;
	m = rp * rp + ip * ip;
	C = t * t;
	d = 4.0 - 4.0 * rp * t + m * C;
	x0 = C / d;
	x1 = 2.0 * C / d;
	x2 = x0;
	C *= m;
	y1 = (8.0 - 2.0 * C) / d;
	y2 = -(4.0 + 4.0 * rp * t + C) / d;

	C = w * 0.5;
	if (!IsLopass) 
		k = -cos(C + 0.5) / cos(C - 0.5);
	else
		k = sin(0.5 - C) / sin(0.5 + C);
	C = k * k;
	d = 1.0 + y1 * k - y2 * C;
	a0 = (x0 - x1 * k + x2 * C) / d;
	a1 = (-2.0 * x0 * k + x1 + x1 * C - 2.0 * x2 * k) / d;
	a2 = (x0 * C - x1 * k + x2) / d;
	b1 = (2.0 * k + y1 + y1 * C - 2.0 * y2 * k) / d;
	b2 = (-C - y1 * k + y2) / d;
	if (!IsLopass)
	{
		a1 = -a1;
		b1 = -b1;
	}
}

void GetChebyshevCoeff(double fc, int poles, double ripple, bool IsLopass, double *a, double *b)
{
	int i;
	double *ta = new double[MAX_POLES + 3];
	double *tb = new double[MAX_POLES + 3];
	double a0, a1, a2, b1, b2, sa, sb, s, g;

	memset(a, 0, sizeof(double) * (MAX_POLES + 3));
	memset(b, 0, sizeof(double) * (MAX_POLES + 3));

	a[2] = 1.0;
	b[2] = 1.0;

	for (int p = 1; p <= poles >> 1; p++)
	{
		GetPairCoeff(fc, IsLopass, ripple, p, poles, a0, a1, a2, b1, b2);
		
		for (i = 0; i <= MAX_POLES + 2; i++)
		{
			ta[i] = a[i];
			tb[i] = b[i];
		}

		for (i = 2; i <= MAX_POLES + 2; i++)
		{
			a[i] = a0 * ta[i] + a1 * ta[i - 1] + a2 * ta[i - 2];
			b[i] = tb[i] - b1 * tb[i - 1] - b2 * tb[i - 2];
		}
	}

	b[2] = 0.0;
	for (i = 0; i <= MAX_POLES; i++)
	{
		a[i] = a[i + 2];
		b[i] = -b[i + 2];
	}

	sa = 0.0;
	sb = 0.0;
	s = 1.0;
	for (i = 0; i <= MAX_POLES; i++)
	{
		if (IsLopass)
		{
			sa += a[i];
			sb += b[i];
		}
		else
		{
			sa += a[i] * s;
			sb += b[i] * s;
		}
		s = -s;
	}

	g = sa / (1 - sb);

	for (i = 0; i <= MAX_POLES; i++)
		a[i] /= g;

	delete tb;
	delete ta;
}

void IIRFilter(SampleData s, int n1, double *x, int n2, double *y)
{
	double sum_x, sum_y, *old_x;
	int i, j, x_h = n1 - 1, x_t = 0;

	if (s.N < n1) return;

	int block_size = s.N / 100;
	if (block_size < 8000) block_size = 8000;

	old_x = new double[n1];

	sum_x = 0.0;
	sum_y = 0.0;

	for (i = 0; i < n1 - 1; i++)
		old_x[i] = s.GetSample(i);

	for (i = n1 - 1; i < s.N; i++)
	{
		old_x[x_h] = s.GetSample(i);
		sum_x = 0.0;
		for (j = 0; j < n1; j++)
			sum_x += old_x[(x_t + n1 - 1 - j) % n1] * x[j];
		sum_y = 0.0;
		for (j = 0; j < n2; j++)
			sum_y += s.GetSample(i - j - 1) * y[j];
		x_h++; x_h %= n1;
		x_t++; x_t %= n1;
		s.SetSample(i, sum_x + sum_y);
		if (!(i % block_size))
			pd->m_Progress.SetPos(int((double)i * 100 / s.N));
	}

	delete old_x;
}

void SetProgress()
{
	pd->m_Progress.SetPos(percentage);
}

void FFTFIRFilter(SampleData s, int n, double *kernel)
{
	int conv_size = 2 * n - 1, i, j, index;
	int N = 1 << (int)ceil(log(conv_size) / log(2));
	int segments = (int)ceil((double)s.N / n);
	double *old_buf, *re_h, *im_h, *re_x, *im_x, tr, ti;
	char *sign, *old_sign;

	percentage = 0;
	old_buf = new double[conv_size + 1];
	re_x = new double[N];
	im_x = new double[N];
	re_h = new double[N];
	im_h = new double[N];

	memset(im_h, 0, sizeof(double) * N);
	memset(re_h, 0, sizeof(double) * N);
	memcpy(re_h, kernel, sizeof(double) * n);
	FFT(N, re_h, im_h);

	memset(old_buf, 0, sizeof(double) * (conv_size + 1));
	if (tf)
	{
		sign = new char[N];
		old_sign = new char[conv_size + 1];
		memset(old_sign, 0, conv_size + 1);
	}

	for (i = 0; i < segments; i++)
	{
		memset(re_x + n, 0, sizeof(double) * (N - n));
		if (tf) memset(sign + n, 0, conv_size - n);
		for (j = 0; j < n; j++)
			if (j + i * n < s.N) 
			{
				re_x[j] = s.GetSample(j + i * n);
				if (tf) 
				{
					sign[j] = 0;
					if (re_x[j])
					{
						if (re_x[j] < 0) sign[j] = 1;
						re_x[j] = log(fabs(re_x[j]));
					}
				}
			}
		memset(im_x, 0, sizeof(double) * N);
		FFT(N, re_x, im_x);
		for (j = 0; j < N; j++)
		{
			tr = re_x[j];
			ti = im_x[j];
			re_x[j] = re_h[j] * tr - im_h[j] * ti;
			im_x[j] = -(re_h[j] * ti + im_h[j] * tr);
		}
		FFT(N, re_x, im_x);
		for (j = 0; j < conv_size; j++)
			re_x[j] /= N;
		if (tf)
		{
			for (j = n; j >= 0; j--)
				sign[j + (n >> 1)] = sign[j];
			memset(sign, 0, n >> 1);
		}
		for (j = 0; j < n; j++)
		{
			index = i * n + j;
			if (index >= s.N) break;
			tr = re_x[j] + old_buf[j + n];
			if (tf)
			{
				tr = exp(tr);
				if (sign[j] + old_sign[j + n]) tr = -tr;
			}
			s.SetSample(index, tr);
		}
		memcpy(old_buf, re_x, sizeof(double) * conv_size);
		if (tf) memcpy(old_sign, sign, conv_size);
		percentage = i * 100 / segments;
		SetProgress();
	}

	if (tf)
	{
		delete sign;
		delete old_sign;
	}
	delete im_h;
	delete re_h;
	delete im_x;
	delete re_x;
	delete old_buf;
}

void HomoFFTFIRFilter(SampleData s, int n, double *kernel)
{
	int conv_size = 2 * n - 1, i, j, index;
	int N = 1 << (int)ceil(log(conv_size) / log(2));
	int N2 = 1 << (int)ceil(log(s.N) / log(2));
	int segments = (int)ceil((double)N2 / n);
	double *old_buf_re, *old_buf_im;
	double *re_h, *im_h, *re_x, *im_x, tr, ti;
	double *re, *im;

	re = new double[N2];
	im = new double[N2];
	for (i = 0; i < s.N; i++)
		re[i] = s.GetSample(i);
	if (s.N < N2) memset(re + s.N, 0, sizeof(double) * (N2 - s.N));
	ForwardFFT(N2, re, im);

	percentage = 0;
	old_buf_re = new double[conv_size + 1];
	old_buf_im = new double[conv_size + 1];
	re_x = new double[N];
	im_x = new double[N];
	re_h = new double[N];
	im_h = new double[N];

	memset(im_h, 0, sizeof(double) * N);
	memset(re_h, 0, sizeof(double) * N);
	memcpy(re_h, kernel, sizeof(double) * n);
	FFT(N, re_h, im_h);

	memset(old_buf_re, 0, sizeof(double) * (conv_size + 1));
	memset(old_buf_im, 0, sizeof(double) * (conv_size + 1));

	for (i = 0; i < segments; i++)
	{
		memset(re_x + n, 0, sizeof(double) * (N - n));
		memset(im_x + n, 0, sizeof(double) * (N - n));
		for (j = 0; j < n; j++)
			if (j + i * n < N2) 
			{
				re_x[j] = re[j + i * n];
				im_x[j] = im[j + i * n];
				if (re_x[j] || im_x[j])
				{
					Complex z(re_x[j], im_x[j]);
					re_x[j] = log(z.Mod());
					im_x[j] = z.Arg();
				}
			}
		FFT(N, re_x, im_x);
		for (j = 0; j < N; j++)
		{
			tr = re_x[j];
			ti = im_x[j];
			re_x[j] = re_h[j] * tr - im_h[j] * ti;
			im_x[j] = re_h[j] * ti + im_h[j] * tr;
		}
		InverseFFT(N, re_x, im_x);
		for (j = 0; j < n; j++)
		{
			index = i * n + j;
			if (index >= N2) break;
			tr = re_x[j] + old_buf_re[j + n];
			ti = im_x[j] + old_buf_im[j + n];
			tr = exp(tr);
			re[index] = tr * cos(ti);
			im[index] = tr * sin(ti);
		}
		memcpy(old_buf_re, re_x, sizeof(double) * conv_size);
		memcpy(old_buf_im, im_x, sizeof(double) * conv_size);
		percentage = i * 100 / segments;
		SetProgress();
	}

	InverseFFT(N2, re, im);
	for (i = 0; i < s.N; i++)
		s.SetSample(i, re[i]);

	delete im_h;
	delete re_h;
	delete im_x;
	delete re_x;
	delete old_buf_re;
	delete old_buf_im;
	delete im;
	delete re;
}

void FIRFilter(SampleData s, int n, double *kernel)
{
	int conv_size = 2 * n - 1, i, j, index, c;
	int segments = (int)ceil((double)s.N / n);
	double *old_buf, *buf, d;
	char *sign, *old_sign;

	percentage = 0;
	buf = new double[conv_size];
	old_buf = new double[conv_size + 1];
	if (tf)
	{
		sign = new char[conv_size];
		old_sign = new char[conv_size + 1];
		memset(old_sign, 0, conv_size + 1);
	}

	memset(old_buf, 0, sizeof(double) * (conv_size + 1));

	for (i = 0; i < segments; i++)
	{
		c = i * n;
		memset(buf + n, 0, sizeof(double) * (conv_size - n));
		if (tf) memset(sign + n, 0, conv_size - n);
		for (j = 0; j < n; j++)
			if (j + c < s.N) 
			{
				buf[j] = s.GetSample(j + c);
				if (tf) 
				{
					sign[j] = 0;
					if (buf[j])
					{
						if (buf[j] < 0) sign[j] = 1;
						buf[j] = log(fabs(buf[j]));
					}
				}
			}
		Convolution(n, buf, n, kernel, buf);
		if (tf)
		{
			for (j = n; j >= 0; j--)
				sign[j + (n >> 1)] = sign[j];
			memset(sign, 0, n >> 1);
		}
		for (j = 0; j < n; j++)
		{
			index = c + j;
			if (index >= s.N) break;
			d = buf[j] + old_buf[j + n];
			if (tf)
			{
				d = exp(d);
				if (sign[j] + old_sign[j + n]) d = -d;
			}
			s.SetSample(index, d);
		}
		memcpy(old_buf, buf, sizeof(double) * conv_size);
		if (tf) memcpy(old_sign, sign, conv_size);
		percentage = i * 100 / segments;
		SetProgress();
	}

	if (tf)
	{
		delete sign;
		delete old_sign;
	}
	delete buf;
	delete old_buf;
}

void MovingAverageFilter(SampleData s, int N)
{
	double new_y, old_y = 0.0, new_x, old_x;
	int i, block_size = 16383;

	if (s.N < N) return;

	old_x = s.GetSample(0);
	for (i = 0; i < N; i++)
		old_y += s.GetSample(i);
	old_y /= N;
	s.SetSample(0, old_y);

	for (i = 1; i < s.N - N; i++)
	{
		new_x = s.GetSample(i + N);
		new_y = old_y + new_x / N - old_x / N;
		old_x = s.GetSample(i);
		s.SetSample(i, new_y);
		old_y = new_y;
		if (!(i & block_size))
			pd->m_Progress.SetPos(int((double)i * 100 / s.N));
	}
}

void RemoveDCOffset(SampleData s)
{
	int i, inc, block_size = (1 << 16) - 1;
	double d = 0.0;

	inc = 1 << s.IsStereo;
	if (s.Is16bit)
	{
		short *data = (short *)s.data, p;
		data += s.Channel;
		for (i = 0; i < s.N; i++, data += inc)
		{
			d += *data;
			if (!(i & block_size))
				pd->m_Progress.SetPos(int((double)i * 50 / s.N));
		}
		p = short(d / s.N);
		data = (short *)s.data;
		data += s.Channel;
		for (i = 0; i < s.N; i++, data += inc)
		{
			*data -= p;
			if (!(i & block_size))
				pd->m_Progress.SetPos(int((double)i * 50 / s.N) + 50);
		}
	}
	else
	{
		unsigned char *data = (unsigned char *)s.data;
		char p;
		data += s.Channel;
		for (i = 0; i < s.N; i++, data += inc)
		{
			d += *data;
			if (!(i & block_size))
				pd->m_Progress.SetPos(int((double)i * 50 / s.N));
		}
		p = char(d / s.N) - 128;
		data = (unsigned char *)s.data;
		for (i = 0; i < s.N; i++, data += inc)
		{
			*data -= p;
			if (!(i & block_size))
				pd->m_Progress.SetPos(int((double)i * 50 / s.N) + 50);
		}
	}
}

void Echo(SampleData s, double delay, double decay, double input_amp, double echo_amp, bool IsMultiple, int smp_sec)
{
	int M = int(delay * smp_sec);
	if (s.N <= M || M < 1) return;
	int i, j, k, l, m, index;
	int chnl = s.Channel, block_size = (1 << 14) - 1;
	double *buf = new double[M];
	double t1, t2, val;
	char *data = s.data;

	for (i = 0; i < M; i++)
		buf[i] = s.GetSample(i);

	for (i = 0; i < M; i++)
		s.SetSample(i, s.GetSample(i) * input_amp);

	m = M;
	l = s.N;
	if (l > (M << 1)) l = (M << 1);
	for (k = 0; k < 2; k++)
	{
		if (s.Is16bit)
		{
			for (i = m; i < l; i++)
			{
				index = (i << s.IsStereo) + chnl;
				j = i % M;
				t1 = buf[j] * decay;
				val = ((short *)data)[index];
				t2 = val;
				val = val * input_amp + t1 * echo_amp;
				if (val > 32767.0) val = 32767.0;
				if (val < -32768.0) val = -32768.0;
				((short *)data)[index] = short(val);
				buf[j] = t2 + t1;
				if (k && !(i & block_size))
					pd->m_Progress.SetPos(int((double)i * 100 / s.N));
			}
		}
		else
		{
			for (i = m; i < l; i++)
			{
				index = (i << s.IsStereo) + chnl;
				j = i % M;
				t1 = buf[j] * decay;
				val = double((unsigned char)data[index] - 128);
				t2 = val;
				val = val * input_amp + t1 * echo_amp;
				if (val > 127.0) val = 127.0;
				if (val < -128.0) val = -128.0;
				data[index] = (unsigned char)(val + 128);
				buf[j] = t1 + t2;
				if (k && !(i & block_size))
					pd->m_Progress.SetPos(int((double)i * 100 / s.N));
			}
		}
		
		if (!IsMultiple)
			for (i = 0; i < M; i++)
				buf[i] = 0.0;
		
		m = (M << 1);
		if (s.N <= m) break;
		l = s.N;
	}

	delete buf;
}

void Reverberation(SampleData s, double InputAmp, double EarlyAmp, double ReverbAmp, double *early, double *ea, double *tail, double PreDelay, double ReverbTime, int smp_sec)
{
	double decay_const = pow(10, -3 / ReverbTime / smp_sec);
	double g[4], d, t, p;
	double *buf1 = new double[smp_sec >> 1];
	double* buf2[4];
	int i, j, bp1 = 0, bp2 = 0, l1, l2;
	int e[3], m[4], block_size = (1 << 12) - 1;

	if (InputAmp < 40) InputAmp = pow(10, -0.1 * InputAmp);
	else InputAmp = 0.0;
	if (EarlyAmp < 40) EarlyAmp = pow(10, -0.1 * EarlyAmp);
	else EarlyAmp = 0.0;
	if (ReverbAmp < 40) ReverbAmp = pow(10, -0.1 * ReverbAmp);
	else ReverbAmp = 0.0;
	l1 = l2 = smp_sec >> 1;
	memset(buf1, 0, sizeof(double) * l1);

	for (i = 0; i < 3; i++)
		e[i] = int(early[i] * smp_sec * 0.001);

	for (i = 0; i < 4; i++)
	{
		m[i] = int((tail[i] + PreDelay) * smp_sec * 0.001);
		g[i] = pow(decay_const, m[i]);
		buf2[i] = new double[smp_sec >> 1];
		memset(buf2[i], 0, sizeof(double) * l2);
	}

	for (i = 0; i < s.N; i++)
	{
		d = s.GetSample(i);
		buf1[bp1] = d;
		for (t = 0.0, j = 0; j < 4; j++)
		{
			p = g[j] * buf2[j][(bp2 - m[j] + l2) % l2];
			buf2[j][bp2] = p + d;
			t += p;
		}
		for (p = 0.0, j = 0; j < 3; j++)
			p += ea[j] * buf1[(bp1 - e[j] + l1) % l1];
		p *= EarlyAmp;
		d *= InputAmp;
		t *= ReverbAmp;
		s.SetSample(i, d + t + p);
		bp1++;
		bp1 %= l1;
		bp2++;
		bp2 %= l2;
		if (!(i & block_size))
			pd->m_Progress.SetPos(int((double)i * 100 / s.N));
	}

	for (i = 0; i < 4; i++)
		delete buf2[i];
	delete buf1;
}

double Square(double x)
{
	return (Period(x) > M_PI ? 0.0 : 1.0);
}

double Triangle(double x)
{
	double p = Period(x);
	if (p < M_PI) return 2.0 * p / M_PI - 1.0;
	p -= M_PI;
	return 1.0 - 2.0 * p / M_PI;	
}

double Saw(double x)
{
	return Period(x) / M_PI - 1.0;
}

double Sin(double x)
{
	return sin(x);
}

void ModSynth(SampleData s, double freq, double phase, double depth, int smp_sec, int waveform, bool IsSynth)
{
	int i, block_size = (1 << 14) - 1;
	char *data = s.data;
	depth = pow(10, -depth * 0.05);
	freq /= smp_sec;
	phase *= M_PI / 180;
	double c = 1 - depth, c2 = 2 * M_PI * freq, val;
	double (*f)(double x), d;

	if (IsSynth)
	{
		if (s.Is16bit)
			depth *= 32767;
		else
			depth *= 127;
		c = 0.0;
	}

	switch (waveform)
	{
		case 0: f = Sin; break;
		case 1: f = Square; break;
		case 2: f = Triangle; break;
		case 3: f = Saw;
	}

	for (i = 0; i < s.N; i++)
	{
		val = s.GetSample(i);
		d = c + f(i * c2 + phase) * depth;
		if (IsSynth) val += d;
		else val *= d;
		s.SetSample(i, val);
		if (!(i & block_size))
			pd->m_Progress.SetPos(int((double)i * 100 / s.N));
	}
}

void AmplitudeModulation(SampleData s, double freq, double phase, double depth, int smp_sec, int waveform)
{
	ModSynth(s, freq, phase, depth, smp_sec, waveform, false);
}

void WaveSynth(SampleData s, double freq, double phase, double depth, int smp_sec, int waveform)
{
	ModSynth(s, freq, phase, depth, smp_sec, waveform, true);
}

void AddNoise(SampleData s, double freq, double val, int smprate)
{
	int n = NOISE_SEG - 1;
	int conv_size = 2 * n - 1, i, j, index;
	int N = 1 << (int)ceil(log(conv_size) / log(2));
	int segments = (int)ceil((double)s.N / n);
	double *old_buf, *re_h, *im_h, *re_x, *im_x, tr, ti, c;

	percentage = 0;
	old_buf = new double[conv_size + 1];
	re_x = new double[N];
	im_x = new double[N];
	re_h = new double[N];
	im_h = new double[N];

	val = pow(10, -val * 0.05);
	if (s.Is16bit) val *= 32767;
	else val *= 127;
	c = val;
	val /= RAND_MAX >> 1;

	memset(re_h, 0, sizeof(double) * N);
	LowPassFIRKernel(n - 1, (double)freq / smprate, true, re_h);
	ForwardFFT(N, re_h, im_h);
	srand((unsigned)time(NULL));

	memset(old_buf, 0, sizeof(double) * (conv_size + 1));

	for (i = 0; i < segments; i++)
	{
		memset(re_x + n, 0, sizeof(double) * (N - n));
		for (j = 0; j < n; j++)
			re_x[j] = val * rand() - c;
		ForwardFFT(N, re_x, im_x);
		for (j = 0; j < N; j++)
		{
			tr = re_x[j];
			ti = im_x[j];
			re_x[j] = re_h[j] * tr - im_h[j] * ti;
			im_x[j] = -(re_h[j] * ti + im_h[j] * tr);
		}
		FFT(N, re_x, im_x);
		for (j = 0; j < conv_size; j++)
			re_x[j] /= N;
		for (j = 0; j < n; j++)
		{
			index = i * n + j;
			if (index >= s.N) break;
			tr = re_x[j] + old_buf[j + n];
			s.SetSample(index, tr + s.GetSample(index));
		}
		memcpy(old_buf, re_x, sizeof(double) * conv_size);
		percentage = i * 100 / segments;
		SetProgress();
	}

	delete im_h;
	delete re_h;
	delete im_x;
	delete re_x;
	delete old_buf;
}

void Fade(SampleData s, int FadeType, bool IsOut)
{
	int i, block_size = (1 << 14) - 1;
	double d, c, c2 = log(0.001);

	for (i = 0; i < s.N; i++)
	{
		d = s.GetSample(i);
		c = (double)i / s.N;
		if (IsOut) c = 1 - c;
		if (!FadeType) d *= 1 - c;
		else d *= exp(c * c2);
		s.SetSample(i, d);
		if (!(i & block_size))
			pd->m_Progress.SetPos(int((double)i * 100 / s.N));
	}
}

void PitchShift(SampleData s, int ShiftBy)
{
	int N = (int)ceil((double)s.N / PITCH_SEG);
	int i, j, index, N2, k;
	double c = pow(2, (double)ShiftBy / 12);
	double *re = new double[PITCH_SEG];
	double *im = new double[PITCH_SEG];
	double *re_x = new double[PITCH_SEG];
	double *im_x = new double[PITCH_SEG];

	if (c == 1.0) return;

	for (i = 0; i < N; i++)
	{
		index = i * PITCH_SEG;
		N2 = s.N - index;
		if (N2 < PITCH_SEG)
		{
			for (j = index; j < s.N; j++)
				s.SetSample(j, 0.0);
			break;
		}
		else
			N2 = PITCH_SEG;
		for (j = 0; j < N2; j++)
			re[j] = s.GetSample(j + index);
		ForwardFFT(PITCH_SEG, re, im);
		memset(re_x, 0, sizeof(double) * ((PITCH_SEG >> 1) + 1));
		memset(im_x, 0, sizeof(double) * ((PITCH_SEG >> 1) + 1));
		for (j = 0; j < PITCH_SEG >> 1; j++)
		{
			k = (int)(c * j);
			re_x[k] = re[j];
			im_x[k] = im[j];
		}
		for (j = 0; j < PITCH_SEG >> 1; j++)
		{
			re_x[PITCH_SEG - j - 1] = re_x[j + 1];
			im_x[PITCH_SEG - j - 1] = -im_x[j + 1];
		}
		InverseFFT(PITCH_SEG, re_x, im_x);
		for (j = 0; j < N2; j++)
			s.SetSample(j + index, re_x[j]);
		pd->m_Progress.SetPos(int((double)index * 100 / s.N));
	}

	delete im_x;
	delete re_x;
	delete im;
	delete re;
}

void Chorus(SampleData s, int ChorusSize, int ChorusDelay, double ModFreq, double ModDepth, double InputAmp, double ChorusAmp, int smprate)
{
	double *copy = new double[s.N], d, c1, c2, c3, c4;
	double phases[5], pos[5], delay;
	int i, j, index, block_size = (1 << 12) - 1;

	for (i = 0; i < s.N; i++)
		copy[i] = s.GetSample(i);

	c1 = 2 * M_PI / RAND_MAX;
	for (i = 0; i < ChorusSize; i++)
	{
		phases[i] = rand() * c1;
		pos[i] = 0.0;
	}

	c1 = ModFreq * 2 * M_PI / smprate;
	c2 = log(2) * ModDepth * 0.01 / 12;
	if (ChorusAmp == 80) c3 = 0.0;
	else c3 = pow(10, -ChorusAmp * 0.05);
	if (InputAmp == 80) c4 = 0.0;
	else c4 = pow(10, -InputAmp * 0.05);
	delay = ChorusDelay * 0.001 / smprate;
	for (i = 0; i < s.N; i++)
	{
		d = 0.0;
		for (j = 0; j < ChorusSize; j++)
		{
			index = int(pos[j] + delay * j);
			if (index >= s.N) break;
			d += copy[index];
			pos[j] += exp(sin(c1 * i + phases[j]) * c2);
		}
		d = d * c3 + copy[i] * c4;
		s.SetSample(i, d);
		if (!(i & block_size))
			pd->m_Progress.SetPos(int((double)i * 100 / s.N));
	}

	delete copy;
}

void Reverse(SampleData s)
{
	int i, inc, block_size = (1 << 16) - 1;

	inc = 1 << s.IsStereo;
	if (s.Is16bit)
	{
		short *data1 = (short *)s.data, *data2 = data1;
		short d;
		data1 += s.Channel;
		if (s.IsStereo)
			data2 += (s.N << 1) - 2 + s.Channel;
		else
			data2 += s.N - 1;
		for (i = 0; i < s.N >> 1; i++, data1 += inc, data2 -= inc)
		{
			d = *data1;
			*data1 = *data2;
			*data2 = d;
			if (!(i & block_size))
				pd->m_Progress.SetPos(int((double)i * 100 / s.N));
		}
	}
	else
	{
		char *data1 = s.data, *data2 = data1;
		char d;
		data1 += s.Channel;
		data2 += s.N - 2 + s.Channel;
		for (i = 0; i < s.N >> 1; i++, data1 += inc, data2 -= inc)
		{
			d = *data1;
			*data1 = *data2;
			*data2 = d;
			if (!(i & block_size))
				pd->m_Progress.SetPos(int((double)i * 100 / s.N));
		}
	}
}

void GetPoleZeroCoeff(double *rp, double *wp, double *r0, double *w0, double *a, double *b, int poles, int zeros, bool Normalize)
{
	int i;
	double *ta = new double[MAX_POLES + 3];
	double *tb = new double[MAX_POLES + 3];
	double a0, a1, a2, b1, b2, g, sa, sb;

	for (i = 0; i <= MAX_POLES + 2; i++)
	{
		a[i] = 0;
		b[i] = 0;
	}

	a[2] = 1.0;
	b[2] = 1.0;

	for (int p = 1; p <= max(poles, zeros); p++)
	{
		i = p - 1;
		a0 = 1.0;
		if (p <= zeros)
		{
			a1 = -2 * r0[i] * cos(w0[i]);
			a2 = r0[i];
		}
		else
		{
			a1 = 0.0;
			a2 = 0.0;
		}
		if (p <= poles)
		{
			b1 = 2 * rp[i] * cos(wp[i]);
			b2 = -rp[i] * rp[i];
		}
		else
		{
			b1 = 0.0;
			b2 = 0.0;
		}
		
		for (i = 0; i <= MAX_POLES + 2; i++)
		{
			ta[i] = a[i];
			tb[i] = b[i];
		}

		for (i = 2; i <= MAX_POLES + 2; i++)
		{
			a[i] = a0 * ta[i] + a1 * ta[i - 1] + a2 * ta[i - 2];
			b[i] = tb[i] - b1 * tb[i - 1] - b2 * tb[i - 2];
		}
	}

	b[2] = 0.0;
	for (i = 0; i <= MAX_POLES; i++)
	{
		a[i] = a[i + 2];
		b[i] = -b[i + 2];
	}

	if (Normalize)
	{
		sa = 0.0;
		sb = 0.0;
		for (i = 0; i <= MAX_POLES; i++)
		{
			sa += a[i];
			sb += b[i];
		}

		g = sa / (1 - sb);

		for (i = 0; i <= MAX_POLES; i++)
			a[i] /= g;
	}

	delete tb;
	delete ta;
}

void GetImpulseResponse(int num_a, double *a, int num_b, double *b, int length, double *data)
{
	double *d = new double[length];
	double *old_x, gain = 0.0;
	double sum_x, sum_y;
	int x_h = num_a - 1, x_t = 0, i, j;

	d[0] = 1.0;
	for (i = 1; i < length; i++)
		d[i] = 0;

	old_x = new double[num_a];
	for (i = 0; i < num_a - 1; i++)
		old_x[i] = d[i];

	for (i = num_a - 1; i < length; i++)
	{
		old_x[x_h] = d[i];
		sum_x = 0.0;
		for (j = 0; j < num_a; j++)
			sum_x += old_x[(x_t + num_a - 1 - j) % num_a] * a[j];
		sum_y = 0.0;
		for (j = 0; j < num_b; j++)
			if (i - j - 1 >= 0)
				sum_y += d[i - j - 1] * b[j + 1];
		x_h++; x_h %= num_a;
		x_t++; x_t %= num_a;
		d[i] = sum_x + sum_y;
	}

	memcpy(data, d, sizeof(double) * length);

	delete old_x;
	delete d;
}

Complex GetTransferFuncValue(Complex z, double *rp, double *wp, double *r0, double *w0, int poles, int zeros)
{
	Complex numerator(1.0);
	Complex denumerator(1.0);
	Complex t;
	int i;

	for (i = 0; i < zeros; i++)
	{
		t.FromPolar(r0[i], w0[i]);
		numerator *= z - t;
	}

	for (i = 0; i < poles; i++)
	{
		t.FromPolar(rp[i], wp[i]);
		denumerator *= z - t;
	}

	return numerator / denumerator;
}

int GetTransformFlag()
{
	return tf;
}

void SamplesToDouble(SampleData s, double *x, double scale)
{
	for (int i = 0; i < s.N; i++)
	{
		x[i] = s.GetSample(i);
		x[i] /= scale;
	}
}

void GetSpectrum(SampleData s, double *spectrum)
{
	int N = s.N, i;
	double *re = new double[N];
	double *im = new double[N];

	SamplesToDouble(s, re, 1);
	SamplesToDouble(s, re, 1);

	for (i = 0; i < N; i++)
		re[i] *= Hamming(N, i);

	ForwardFFT(N, re, im);

	for (i = 0; i < N >> 1; i++)
	{
		re[i] /= N;
		im[i] /= N;
		spectrum[i] = sqrt(re[i] * re[i] + im[i] * im[i]);
	}

	delete im;
	delete re;
}

double LevelTo_dB(double level, double cmp_level, double min_dB)
{
	double res;

	if (level)
		res = 20 * log10(level / cmp_level);
	else
		return min_dB;

	if (res < min_dB) return min_dB;

	return res;
}

int GetEqualizerFreq(int n, int HalfRate)
{
	if (n < 5) return freqs[0][n + 1];

	for (int i = 0; i < 3; i++)
		if (HalfRate == freqs[i + 1][0]) return freqs[i + 1][n - 4];

	return n * ((HalfRate - 1000) / 1000) / 5 + 1000;
}

int GetSampleRate(int index)
{
	return smp_rates[index];
}

void SetProgressDlg(CProgressDlg *pd)
{
	::pd = pd;
}

void SetTransformFlag(int tf)
{
	::tf = tf;
}
